/*
	*
	* This file is a part of CoreFM.
	* A file manager for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* ( at your option ) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <QApplication>
#include <QClipboard>
#include <QScroller>
#include <QMenu>
#include <QListWidgetItem>

#include <cprime/corexdg.h>
#include <cprime/themefunc.h>

#include <csys/storageinfo.h>

#include "filedialog.h"
#include "folderview.hpp"
#include "corefm.h"
#include "ui_corefm.h"

// Init moveItems
bool corefm::moveItems = false;

corefm::corefm( QString path, QWidget *parent )
	: QWidget( parent )
	, ui( new Ui::corefm )
	, isSplit(false)
	, clipBoard(QApplication::clipboard())
	, mimeHandler(CoreXdgMime::instance())
	, sMgr(StorageManager::instance())
{
    ui->setupUi( this );

    view1 = new FolderView( this );
    view2 = new FolderView( this );
    view1->setObjectName("view1");
    view2->setObjectName("view2");
    view1->setFocusPolicy(Qt::StrongFocus);
    view2->setFocusPolicy(Qt::StrongFocus);

    QScroller::grabGesture(ui->pageTrash, QScroller::LeftMouseButtonGesture);
    QScroller::grabGesture(ui->driveView->viewport(), QScroller::LeftMouseButtonGesture);

    QVariant OvershootPolicy = QVariant::fromValue<QScrollerProperties::OvershootPolicy>(QScrollerProperties::OvershootAlwaysOff);
    QScrollerProperties ScrollerProperties = QScroller::scroller(ui->driveView->viewport())->scrollerProperties();
    ScrollerProperties.setScrollMetric(QScrollerProperties::VerticalOvershootPolicy, OvershootPolicy);
    ScrollerProperties.setScrollMetric(QScrollerProperties::HorizontalOvershootPolicy, OvershootPolicy);
    QScroller::scroller(ui->driveView->viewport())->setScrollerProperties(ScrollerProperties);

    ui->splitter->addWidget( view1 );
    ui->splitter->addWidget( view2 );

    path = QFileInfo( path ).absoluteFilePath();

    loadSettings();
    startSetup();
	setupActions();
	makeConnections();

    view1->loadAddress( path );
	view2->loadAddress( sm->value( "CoreFM", "StartupPath" ) );
	view2->hide();

	view1->activateWindow();
	view1->setFocus();
}

corefm::~corefm()
{
    delete ui;
}

/**
 * @brief Setup ui elements
 */
void corefm::startSetup()
{
    // all toolbuttons icon size in sideView
    for ( QToolButton *b : ui->sideView->findChildren<QToolButton *>() ) {
        if ( b ) {
            b->setIconSize( sideViewIconSize );
        }
    }

    // all toolbuttons icon size in toolsBar
    for ( QToolButton *b : ui->toolsBar->findChildren<QToolButton *>() ) {
        if ( b ) {
            b->setIconSize( toolBarIconSize );
        }
    }

    // all toolbuttons icon size in topBar
    for ( QToolButton *b : ui->topBar->findChildren<QToolButton *>() ) {
        if ( b ) {
            b->setIconSize( toolBarIconSize );
        }
    }

    // set stylesheet from style.qrc
    ui->sideView->setStyleSheet( CPrime::ThemeFunc::getSideViewStyleSheet() );
    ui->sideView->setFixedWidth( sideViewSize );

    view1->setIconSize( listViewIconSize );
    view2->setIconSize( listViewIconSize );

    ui->toolsBar->setVisible( 0 );

    ui->menu->setVisible( 0 );
    ui->appTitle->setAttribute( Qt::WA_TransparentForMouseEvents );
    ui->appTitle->setFocusPolicy( Qt::NoFocus );
    this->resize( 800, 500 );

    connect( ui->driveView, &QListWidget::itemClicked, this, &corefm::openSelecteDrive);

    if(CPrime::FileUtils::exists( QDir::home().filePath( "Downloads" ) ))
        ui->downloads->setVisible(1);
    else
        ui->downloads->setVisible(0);

    if(CPrime::FileUtils::exists( QDir::home().filePath( "Desktop" ) ))
        ui->desktop->setVisible(1);
    else
        ui->desktop->setVisible(0);

    if ( uiMode != 0 ) {
        this->setWindowState( Qt::WindowMaximized );

        if ( uiMode == 2 ){
            ui->menu->setIconSize( toolBarIconSize );

            connect( ui->menu, &QToolButton::clicked, this, &corefm::showSideView );
            connect( ui->appTitle, &QToolButton::clicked, this, &corefm::showSideView );

            ui->sideView->setVisible( 0 );
            ui->menu->setVisible( 1 );

            ui->sizeLbl->setVisible(0);
            ui->spaceLbl->setVisible(0);
        }
    }
}

/**
 * @brief Loads application settings
 */
void corefm::loadSettings()
{
    sideViewSize = sm->value( "CoreApps", "SideViewSize" );
    sideViewIconSize = sm->value( "CoreApps", "SideViewIconSize" );
    toolBarIconSize = sm->value( "CoreApps", "ToolBarIconSize" );
    listViewIconSize = sm->value("CoreApps", "ListViewIconSize");
    uiMode = sm->value("CoreApps", "UIMode");
    showHidden = sm->value( "CoreFM", "ShowHidden" );
    viewMode = static_cast<int>( sm->value( "CoreFM", "ViewMode" ) );
    showThumb = sm->value( "CoreFM", "ShowThumb" );
}

void corefm::closeEvent(QCloseEvent *event)
{
    QString currentFile;

    if ( currentView()->selection().count() )
        currentFile = currentView()->selection().value( 0 );

    else
        currentFile = currentView()->rootPath();

    CPrime::InfoFunc::saveToRecent("corefm", currentFile);
    event->accept();
}

void corefm::showSideView()
{
    if ( ui->sideView->isVisible() )
        ui->sideView->setVisible( 0 );
    else
        ui->sideView->setVisible( 1 );
}

void corefm::setupActions() {

    newFolderAct = new QAction( QIcon::fromTheme( "folder-new" ), "New Folder", this );
	newFolderAct->setShortcut( tr( "F10" ) );
	connect( newFolderAct, SIGNAL( triggered() ), this, SLOT( newFolder() ) );

	newFileAct = new QAction( QIcon::fromTheme( "document-new" ), "New Text", this );
	newFileAct->setShortcut( tr( "Shift+F10" ) );
	connect( newFileAct, SIGNAL( triggered() ), this, SLOT( newFile() ) );

	showHiddenAct = new QAction( QIcon::fromTheme( "password-show-on" ), "Hidden", this );
	showHiddenAct->setShortcut( tr( "Ctrl+H" ) );
	connect( showHiddenAct, SIGNAL( triggered() ), this, SLOT( toggleHidden() ) );
	showHiddenAct->setCheckable( true );
	showHiddenAct->setChecked( showHidden );

	previewAct = new QAction( QIcon::fromTheme( "bitmap-trace" ), "Toggle Previews", this );
	previewAct->setShortcut( tr( "Ctrl+Shift+I" ) );
	connect( previewAct, SIGNAL( triggered() ), this, SLOT( togglePreviews() ) );
	previewAct->setCheckable( true );
	previewAct->setChecked( showThumb );

    termAct = new QAction( QIcon::fromTheme( "utilities-terminal" ), "Open Terminal Here", this );
	termAct->setShortcut( tr( "F4" ) );
	connect( termAct, SIGNAL( triggered() ), this, SLOT( openTerminal() ) );

	cutAct = new QAction( QIcon::fromTheme( "edit-cut" ), "Cut", this );
	cutAct->setShortcut( tr( "Ctrl+X" ) );
	connect( cutAct, SIGNAL( triggered() ), this, SLOT( prepareMove() ) );

	copyAct = new QAction( QIcon::fromTheme( "edit-copy" ), "Copy", this );
	copyAct->setShortcut( tr( "Ctrl+C" ) );
	connect( copyAct, SIGNAL( triggered() ), this, SLOT( prepareCopy() ) );

	pasteAct = new QAction( QIcon::fromTheme( "edit-paste" ), "Paste", this );
	pasteAct->setShortcut( tr( "Ctrl+V" ) );
	connect( pasteAct, SIGNAL( triggered() ), this, SLOT( paste() ) );

	openAct = new QAction( QIcon::fromTheme( "document-open" ), "Open", this );
	openAct->setShortcuts( QList<QKeySequence>() << tr( "Enter" ) << tr( "Return" ) );
	connect( openAct, SIGNAL( triggered() ), this, SLOT( loadAddress() ) );

	refreshAct = new QAction( QIcon::fromTheme( "view-refresh" ), "Refresh", this );
	refreshAct->setShortcut( tr( "F5" ) );
	connect( refreshAct, SIGNAL( triggered() ), this, SLOT( refresh() ) );

	selectAllAct = new QAction( QIcon::fromTheme( "edit-select-all" ), "Select All", this );
	selectAllAct->setShortcut( tr( "Ctrl+A" ) );
	connect( selectAllAct, SIGNAL( triggered() ), this, SLOT( selectAll() ) );

	renameAct = new QAction( QIcon::fromTheme( "edit-rename" ), "Rename", this );
	renameAct->setShortcut( tr( "F2" ) );
	connect( renameAct, SIGNAL( triggered() ), this, SLOT( rename() ) );

	batchRenameAct = new QAction( QIcon::fromTheme( "edit-rename" ), "Batch Rename", this );
	batchRenameAct->setShortcut( tr( "Shift+F2" ) );
    connect( batchRenameAct, SIGNAL( triggered() ), this, SLOT( openRenamer() ) );

	trashAct = new QAction( QIcon::fromTheme( "edit-delete" ), "Move to Trash", this );
	trashAct->setShortcut( tr( "Delete" ) );
	connect( trashAct, SIGNAL( triggered() ), this, SLOT( trashSelected() ) );

	deleteAct = new QAction( QIcon::fromTheme( "edit-delete" ) , "Delete", this );
	deleteAct->setShortcut( tr( "Shift+Delete" ) );
	connect( deleteAct, SIGNAL( triggered() ), this, SLOT( deleteSelected() ) );

    propsAct = new QAction( QIcon::fromTheme( "document-properties" ), "Properties", this );
	propsAct->setShortcut( tr( "Alt+Return" ) );
    connect( propsAct, SIGNAL( triggered() ), this, SLOT( showProperties() ) );

    backAct = new QAction( QIcon::fromTheme( "go-next" ), "Previous", this );
	backAct->setShortcut( tr( "Alt+Left" ) );
    connect( backAct, SIGNAL( triggered() ), this, SLOT( goBack() ) );

    openWithAct = new QAction( QIcon::fromTheme( "quickopen-file" ), "Open With", this );
    openWithAct->setShortcut( tr( "Ctrl+Return" ) );
    connect( openWithAct, SIGNAL( triggered() ), this, SLOT( selectApp() ) );

    upAct = new QAction( QIcon::fromTheme( "go-previous" ), "Parent dir", this );
    upAct->setShortcut( tr( "Backspace" ) );
	connect( upAct, SIGNAL( triggered() ), this, SLOT( goUp() ) );

	splitAct = new QAction( QIcon::fromTheme( "view-split-left-right" ), "Toggle Split View", this );
	connect( splitAct, SIGNAL( triggered() ), this, SLOT( toggleSplit() ) );
	splitAct->setCheckable( true );
	splitAct->setChecked( false );

	openInSplitAct = new QAction( QIcon::fromTheme( "view-split-left-right" ), "Open folder in SplitView", this );
	openInSplitAct->setShortcut( tr( "Ctrl+Shift+Return" ) );
	connect( openInSplitAct, SIGNAL( triggered() ), this, SLOT( openInSplitView() ) );

    viewChangeAct = new QAction( QIcon::fromTheme( "view-list" ), "Change View", this );
    viewChangeAct->setShortcut( tr( "Ctrl+Shift+V" ) );
	switch( viewMode ) {
		case 0: {
            viewChangeAct->setIcon( QIcon::fromTheme( "view-list-icons" ) );
			break;
		}
        case 1: {
            viewChangeAct->setIcon( QIcon::fromTheme( "view-list-details" ) );
			break;
		}
		case 2: {
            viewChangeAct->setIcon( QIcon::fromTheme( "view-list-text" ) );
			break;
		}
    }
    connect( viewChangeAct, SIGNAL( triggered() ), this, SLOT( nextView() ) );

	newWindowAct = new QAction( QIcon::fromTheme( "window-new" ), "New Window", this );
	// extractAct->setShortcut( tr( "" ) );
	connect( newWindowAct, SIGNAL( triggered() ), this, SLOT( newWindow() ) );

	sortNameAct = new QAction( "Name", this );
	sortNameAct->setCheckable( true );
	sortNameAct->setChecked( true );

	sortSizeAct = new QAction( "Size", this );
	sortSizeAct->setCheckable( true );
	sortSizeAct->setChecked( false );

	sortDateAct = new QAction( "Date", this );
	sortSizeAct->setCheckable( true );
	sortSizeAct->setChecked( false );

	sortOrderAct = new QAction( "Ascending", this );
	sortOrderAct->setCheckable( true );
	sortOrderAct->setChecked( true );

	QActionGroup *sortGroup = new QActionGroup( this );
	sortGroup->setExclusive( true );
	sortGroup->addAction( sortNameAct );
	sortGroup->addAction( sortSizeAct );
	sortGroup->addAction( sortDateAct );

	connect( sortGroup, SIGNAL( triggered( QAction * ) ), this, SLOT( setSortType( QAction * ) ) );
	connect( sortOrderAct, SIGNAL( triggered() ), this, SLOT( setSortOrder() ) );

	treeWriteAct = new QAction( QIcon::fromTheme( "document-new" ), "Save Dir Tree as text", this );
	connect( treeWriteAct, SIGNAL( triggered() ), this, SLOT( writeTree() ) );

	metadataAct = new QAction( QIcon::fromTheme( "document-properties" ), "Open Metadata viewer", this );
	connect( metadataAct, SIGNAL( triggered() ), this, SLOT( openMetadataViewer() ) );

	searchAct = new QAction( QIcon::fromTheme( "edit-find" ), "Search", this );
	connect( searchAct, SIGNAL( triggered() ), this, SLOT( openSearch() ) );

    ui->back->setDefaultAction( backAct );
    ui->up->setDefaultAction( upAct );
    ui->cut->setDefaultAction( cutAct );
    ui->copy->setDefaultAction( copyAct );
    ui->paste->setDefaultAction( pasteAct );
    ui->trashit->setDefaultAction( trashAct );
    ui->properties->setDefaultAction( propsAct );
    ui->newfolder->setDefaultAction( newFolderAct );
    ui->newtext->setDefaultAction( newFileAct );
    ui->showHidden->setDefaultAction( showHiddenAct );
    ui->showthumb->setDefaultAction( previewAct );
    ui->terminal->setDefaultAction( termAct );
    ui->split->setDefaultAction( splitAct );
    ui->views->setDefaultAction( viewChangeAct );
    ui->selectAll->setDefaultAction( selectAllAct );
    ui->openWith->setDefaultAction( openWithAct );
    ui->search->setDefaultAction( searchAct );
    ui->rename->setDefaultAction( renameAct );

	addAction( newFolderAct );
	addAction( newFileAct );
	addAction( showHiddenAct );
	addAction( previewAct );
	addAction( termAct );
	addAction( cutAct );
	addAction( copyAct );
	addAction( pasteAct );
	addAction( openAct );
	addAction( refreshAct );
	addAction( selectAllAct );
	addAction( renameAct );
	addAction( trashAct );
	addAction( deleteAct );
	addAction( propsAct );
    addAction( batchRenameAct );
	addAction( newWindowAct );
	addAction( sortNameAct );
	addAction( sortSizeAct );
	addAction( sortDateAct );
	addAction( sortOrderAct );
	addAction( backAct );
    addAction( upAct );
	addAction( splitAct );
	addAction( openInSplitAct );
	addAction( viewChangeAct );
    addAction( treeWriteAct );
	addAction( metadataAct );
	addAction( searchAct );
//	addAction( zoomInAct );
//	addAction( zoomOutAct );
};

void corefm::makeConnections() {

	connect( view1, SIGNAL( contextMenu( const QPoint & ) ), this, SLOT( contextMenu( const QPoint & ) ) );
	connect( view2, SIGNAL( contextMenu( const QPoint & ) ), this, SLOT( contextMenu( const QPoint & ) ) );

    connect( view1, SIGNAL( updateVarious() ), this, SLOT( updateVarious() ) );
    connect( view2, SIGNAL( updateVarious() ), this, SLOT( updateVarious() ) );

	connect( view1, SIGNAL( openFile( QString ) ), this, SLOT( openInApp( QString ) ) );
	connect( view2, SIGNAL( openFile( QString ) ), this, SLOT( openInApp( QString ) ) );

//    connect( view1, &FolderView::updateAddress, ui->pathLbl, &QLabel::setText );
//    connect( view2, &FolderView::updateAddress, ui->pathLbl, &QLabel::setText );
};

void corefm::on_tools_clicked()
{
    // first hide the sideview to get more spaces
    if (uiMode == 2){
        ui->sideView->setVisible( 0 );
    }

    if( ui->toolsBar->isVisible() )
        ui->toolsBar->hide();
    else
        ui->toolsBar->show();

}

FolderView* corefm::currentView() {

	if ( not view2->isVisible() )
		return view1;

    if (view2->hasFocus())
        return view2;
    else
        return view1;
};

void corefm::newFolder() {

	QString path = currentView()->rootPath();

    filedialog *dlg = new filedialog( filedialog::NewFolder, path, QString(), this );
    dlg->exec();
};

void corefm::newFile() {

	QString path = currentView()->rootPath();

    filedialog *dlg = new filedialog( filedialog::NewFile, path, QString(), this );
    dlg->exec();
};

void corefm::toggleHidden() {

	showHidden = not showHidden;
	sm->setValue( "CoreFM", "ShowHidden", showHidden );
	currentView()->refresh();
};

void corefm::togglePreviews() {

	showThumb = not showThumb;
	sm->setValue( "CoreFM", "ShowThumb", showThumb );
    currentView()->refresh();
};

void corefm::openTerminal() {

	QStringList selectedList = currentView()->selection();

	if ( selectedList.count() ) {
		QFileInfo file( selectedList.value( 0 ) );
		if( file.isDir() ){
            CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::Terminal, selectedList.value( 0 ), "", this );
			return;
		} else{
            CPrime::InfoFunc::messageEngine( "CoreFM", "corefm", "Can't open terminal here", "Please select a folder to open terminal there.", this );
			return;
		}
	}

    CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::Terminal, currentView()->rootPath(), "", this );
};

void corefm::prepareMove() {

	QStringList selectedItems = currentView()->selection();
	if ( not selectedItems.count() )
		return;

	moveItems = true;

	QList<QUrl> urlList;
	QByteArray uriList;
	foreach( QString item, selectedItems ) {
		QUrl url = QUrl::fromLocalFile( item );
		urlList << url;
		uriList += QString(url.toString() + "\n").toLocal8Bit();
	}

	QMimeData *mData = new QMimeData();
	mData->setUrls( urlList );

	/* GNOME, LXDE, XFCE: Taken from libfm-qt */
	mData->setData( QStringLiteral( "x-special/gnome-copied-files" ), QByteArrayLiteral( "cut\n" ) + uriList );

	/* KDE: Taken from libfm-qt */
	mData->setData( QStringLiteral( "text/uri-list" ), uriList.replace( "\n", "\r\n" ) );
	mData->setData( QStringLiteral( "application/x-kde-cutselection" ), QByteArrayLiteral( "1" ) );

	clipBoard->setMimeData( mData );

	currentView()->clearSelection();
};

void corefm::prepareCopy() {

	if ( not currentView()->selection().count() )
		return;

	moveItems = false;

	QList<QUrl> urlList;
	foreach( QString item, currentView()->selection() )
		urlList << QUrl::fromLocalFile( item );

	QMimeData *mData = new QMimeData();
	mData->setUrls( urlList );

	clipBoard->setMimeData( mData );

	currentView()->clearSelection();
};

void corefm::paste() {

	const QMimeData *mimeData = clipBoard->mimeData();
	bool shouldCutKde = ( mimeData->data( "application/x-kde-cutselection" ).toInt() ? true : false );
	bool shouldCutGnome = ( mimeData->data( "x-special/gnome-copied-files" ).startsWith( "cut\n" ) ? true : false );

	if ( mimeData->hasUrls() ) {
		CoreProcess::Process *process = new CoreProcess::Process;
		process->sourceDir = CPrime::FileUtils::dirName( mimeData->urls().at( 0 ).toLocalFile() );
		process->targetDir = currentView()->rootPath();

		if ( not process->sourceDir.endsWith( "/" ) )
			process->sourceDir += "/";

		QStringList srcList;
		foreach( QUrl url, mimeData->urls() ) {
			QString file = url.toLocalFile();
			if ( CPrime::FileUtils::exists( file ) )
				srcList << file.replace( process->sourceDir, "" );
		}

		if ( not srcList.count() )
			return;

		if ( moveItems or shouldCutKde or shouldCutGnome )
			process->type = CoreProcess::Move;

		else
			process->type = CoreProcess::Copy;

		IODialog *pasteDlg = new IODialog( srcList, process );
		pasteDlg->show();
	}

	clipBoard->setMimeData( new QMimeData() );
	pasteAct->setVisible( false );
};

void corefm::refresh() {

	if ( view1->hasFocus() )
		view1->refresh();

	else
		view2->refresh();
};

void corefm::selectAll() {

	currentView()->selectAll();
};

void corefm::rename() {

	QString path;
	QStringList selection;

     filedialog *dlg = new filedialog( filedialog::Rename, currentView()->rootPath(), currentView()->selection().value( 0 ), this );
     dlg->exec();
};

void corefm::openRenamer() {

    CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::Renamer, currentView()->selection().value( 0 ), "", this );
};

void corefm::openMetadataViewer() {

    CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::MetadataViewer, currentView()->selection().value( 0 ), "", this );
}

void corefm::trashSelected() {

	CPrime::TrashManager::moveToTrash( currentView()->selection() );
	ui->pageTrash->refresh();
};

void corefm::deleteSelected() {

	CPrime::TrashManager::deleteFileTotally( currentView()->selection(), false );
};

void corefm::showProperties() {

	QString currentFile;

	if ( currentView()->selection().count() )
		currentFile = currentView()->selection().value( 0 );

	else
		currentFile = currentView()->rootPath();

	propertiesw *props = new propertiesw( currentFile, this );
	props->show();
};

void corefm::goBack()
{
    if( ui->stackedWidget->currentIndex() == 1 ) {
        ui->stackedWidget->setCurrentIndex( 0 );
		ui->toolsBar->setEnabled( true );
	}

    currentView()->goBack();
};

void corefm::goForward()
{
	if( ui->stackedWidget->currentIndex() == 1 ) {
        ui->stackedWidget->setCurrentIndex( 0 );
		ui->toolsBar->setEnabled( true );
	}

    currentView()->goForward();
};

void corefm::goUp()
{
	if( ui->stackedWidget->currentIndex() == 1 ) {
        ui->stackedWidget->setCurrentIndex( 0 );
		ui->toolsBar->setEnabled( true );
	}

    currentView()->goUp();
};

void corefm::openSearch() {

	QStringList selectedList = currentView()->selection();

	if ( selectedList.count() ) {
		QFileInfo file( selectedList.value( 0 ) );
		if ( file.isDir() )
            CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::SearchApp, selectedList.value( 0 ), "", this );

		else
            CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::SearchApp, CPrime::FileUtils::dirName( selectedList.value( 0 ) ), "", this );

		return;
	}

    CPrime::AppOpenFunc::defaultAppEngine( CPrime::Category::SearchApp, currentView()->rootPath(), "", this );
};

void corefm::writeTree() {

	QString fileName = QFileDialog::getSaveFileName(
		this,
		"CoreFM - Save Directory Tree As",
		currentView()->rootPath(),
		"All Files ( *.* )"
	 );

	if ( not fileName.count() )
		return;

	QString path;
	if ( not currentView()->selection().count() )
		path = currentView()->rootPath();

	else
		path = currentView()->selection().value( 0 );

	TreeWriter *writer = new TreeWriter();
	writer->writeTree( path, fileName );
};

void corefm::nextView() {

    viewMode = ( viewMode + 1 ) % 3;

    /* IconView */
	if ( viewMode == 0 )
        viewChangeAct->setIcon( QIcon::fromTheme( "view-list-icons" ) );

	/* TileView */
	else if ( viewMode == 1 )
        viewChangeAct->setIcon( QIcon::fromTheme( "view-list-details" ) );

	/* DetailsView */
	else
        viewChangeAct->setIcon( QIcon::fromTheme( "view-list-text" ) );

	view1->refreshView( viewMode );
	view2->refreshView( viewMode );
};

void corefm::updateVarious() {

    ui->pathLbl->setText(QFileInfo(currentView()->rootPath()).fileName() );
	setWindowTitle( "CoreFM - " + CPrime::FileUtils::baseName( currentView()->rootPath() ) );

	ui->spaceLbl->setText( QString( "%1 free" ).arg( CPrime::FileFunc::formatSize( QStorageInfo( currentView()->rootPath() ).bytesAvailable() ) ) );

	QStringList selected = currentView()->selection();

	/* Some selection */
	if ( selected.count() ) {

        cutAct->setVisible( true );
        ui->cut->setVisible( true );

        copyAct->setVisible( true );
        ui->copy->setVisible( true );

		trashAct->setVisible(true);
        ui->trashit->setVisible(true);

		/* Multiple nodes selected */
        if ( selected.count() > 1 ) {
			quint64 folders = 0, files = 0;
			for( QString node: selected ) {
				if ( CPrime::FileUtils::isDir( node ) )
					folders++;

				else
					files++;
			}

			QString text = "";
			if ( folders )
				text += QString( "%1 folder%2%3" ).arg( folders ).arg( folders > 1 ? "s" : "" ).arg( files > 0 ? ", " : "" );

			if ( files )
				text += QString( "%1 file%2" ).arg( files ).arg( files > 1 ? "s" : "" );

			ui->sizeLbl->setText( text );

            ui->newtext->setVisible(false);
            ui->newfolder->setVisible(false);
            ui->search->setVisible(false);
            renameAct->setVisible( false );
            ui->rename->setVisible( false );
            propsAct->setVisible( false );
            ui->properties->setVisible( false );
		}

		/* Single node selected */
		else {
			quint64 count = 0;
			if ( CPrime::FileUtils::isDir( selected.at( 0 ) ) ) {
				struct dirent **fileList;
				count = scandir( selected.at( 0 ).toStdString().c_str(), &fileList, nullptr, nullptr ) - 2;
				ui->sizeLbl->setText( QString( "%1 item%2" ).arg( count ).arg( count > 1 ? "s" : "" ) );
			}

			else {
				count = CPrime::FileUtils::getFileSize( selected.at( 0 ) );
				ui->sizeLbl->setText( CPrime::FileFunc::formatSize( count ) );
			}

            ui->newtext->setVisible(true);
            ui->newfolder->setVisible(true);
            ui->search->setVisible(true);
            renameAct->setVisible( true );
            ui->rename->setVisible( true );
            propsAct->setVisible( true );
            ui->properties->setVisible( true );
            openWithAct->setVisible( true );
            ui->openWith->setVisible( true );
		}
	}

	/* No selection */
	else {
		cutAct->setVisible( false );
        ui->cut->setVisible( false );
		copyAct->setVisible( false );
        ui->copy->setVisible( false );
        trashAct->setVisible(false);
        ui->trashit->setVisible(false);

		renameAct->setVisible( false );
        ui->rename->setVisible( false );
        openWithAct->setVisible( false );
        ui->openWith->setVisible( false );
		propsAct->setVisible( true );
        ui->properties->setVisible( true );

		/* The subtraction of 2 is to account for . and .. */
		struct dirent **fileList;
		quint64 count = scandir( currentView()->rootPath().toStdString().c_str(), &fileList, nullptr, nullptr ) - 2;

		if ( count )
			ui->sizeLbl->setText( QString( "%1 item%2" ).arg( count ).arg( count > 1 ? "s" : "" ) );

		else
			ui->sizeLbl->setText( "Empty" );
	}

	if ( currentView()->canGoBack() )
		backAct->setEnabled( true );
	else
		backAct->setDisabled( true );

	/* Clipboard */
    if ( clipBoard->mimeData()->hasUrls() ){
        pasteAct->setVisible( true );
        ui->paste->setVisible( true );
    }
    else{
        pasteAct->setVisible( false );
        ui->paste->setVisible( false );
    }

	/* Up action */
	if ( currentView()->canGoUp() )
		upAct->setEnabled( true );

	else
		upAct->setDisabled( true );
};

void corefm::openInApp( QString path ) {

	if ( path.count() ) {
		CPrime::AppOpenFunc::appOpenEngine( path, this );
        return;
	}

	QAction *action = dynamic_cast<QAction *>( sender() );

	if ( action ) {
		CoreDesktopFile app = action->data().value<CoreDesktopFile>();
		app.startApplicationWithArgs( currentView()->selection() );
	}
};

void corefm::selectApp() {

	// Select application in the dialog
	ApplicationDialog *dialog = new ApplicationDialog( this );

	if ( dialog->exec() ) {
		if ( dialog->getCurrentLauncher().compare( "" ) != 0 ) {
            CPrime::AppOpenFunc::systemAppOpener( dialog->getCurrentLauncher(), currentView()->selection()  );

			// Function from utilities.cpp
			CPrime::InfoFunc::saveToRecent( dialog->getCurrentLauncher(), currentView()->selection().value( 0 ) );
		}
	}
};

void corefm::setSortType( QAction *act ) {

	QString text = act->text();
	bool asc = sortOrderAct->isChecked();

	if ( text == "Name" ) {
		view1->sort( 0, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
		view2->sort( 0, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
	}

	else if ( text == "Size" ) {
		view1->sort( 1, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
		view2->sort( 1, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
	}

	else if ( text == "Date" ) {
		view1->sort( 3, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
		view2->sort( 3, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
	}
};

void corefm::setSortOrder() {

	bool asc = sortOrderAct->isChecked();

	if ( sortNameAct->isChecked() ) {
		view1->sort( 0, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
		view2->sort( 0, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
	}

	else if ( sortSizeAct->isChecked() ) {
		view1->sort( 1, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
		view2->sort( 1, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
	}

	else if ( sortDateAct->isChecked() ) {
		view1->sort( 3, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
		view2->sort( 3, ( asc ? Qt::AscendingOrder : Qt::DescendingOrder ) );
	}
};

void corefm::newWindow( QString path ) {

	corefm *cfm = new corefm( path );

    if ( uiMode != 0 )
		cfm->showMaximized();

	else
		cfm->show();
};

void corefm::openInSplitView() {

	QString path;

	QAction *openAct = qobject_cast<QAction *>( sender() );
	if ( openAct )
		path = currentView()->selection().value( 0 );

	else
		return;

	view2->loadAddress( path );

    if ( not view2->isVisible() ) {
		view2->show();
		view2->activateWindow();
		view2->setFocus();
	}
};

void corefm::loadAddress( QString path ) {

	QAction *openAct = qobject_cast<QAction *>( sender() );

    if ( openAct )
		currentView()->loadAddress( currentView()->selection().value( 0 ) );
    else {
        // first hide the sideview to get more spaces
        if (uiMode == 2){
            ui->sideView->setVisible( 0 );
        }

        if( ui->stackedWidget->currentIndex() == 1 ||  ui->stackedWidget->currentIndex() == 2 ) {
			ui->stackedWidget->setCurrentIndex( 0 );
		}

		if (ui->stackedWidget->currentIndex() == 0) {
			ui->toolsBar->setEnabled(true);
		}

        currentView()->loadAddress( path );
    }
};

void corefm::toggleSplit() {

    if ( view2->isVisible() ) {
		view2->hide();
		view1->activateWindow();
		view1->setFocus();
	}

	else {
		view2->show();
		view1->activateWindow();
		view1->setFocus();
	}
};

void corefm::on_shareIt_clicked()
{
    if ( currentView()->selection().count() ) {
        shareit::shareFiles( this, currentView()->selection() );
    } else {
        CPrime::InfoFunc::messageEngine( "CoreFM", "corefm", "Warning!!!", "Nothing Selected.\nSelect some files to share.", this );
    }
}

void corefm::on_pinIt_clicked()
{
    QStringList selectedItems = currentView()->selection();

    if ( selectedItems.count() ) {
        Q_FOREACH( QString path, selectedItems )
            pinit::callPinIt( this, path );
    }

    else {

         pinit::callPinIt( this, currentView()->rootPath() );
    }
}

void corefm::on_trash_clicked()
{
    // first hide the sideview to get more spaces
    if (uiMode == 2){
        ui->sideView->setVisible( 0 );
    }

    if( ui->stackedWidget->currentIndex() == 1 ) {
        ui->stackedWidget->setCurrentIndex( 0 );
		ui->toolsBar->setEnabled( true );
	}

    else {
        ui->stackedWidget->setCurrentIndex( 1 );
		ui->toolsBar->setDisabled( true );
	}
}

void corefm::on_home_clicked()
{
    loadAddress( QDir::homePath() );
}

void corefm::on_desktop_clicked()
{
    loadAddress( QDir::home().filePath( "Desktop" ) );
}

void corefm::on_downloads_clicked()
{
    loadAddress( QDir::home().filePath( "Downloads" ) );
}

void corefm::on_root_clicked()
{
    loadAddress( "/" );
}

void corefm::on_drives_clicked()
{
    // first hide the sideview to get more spaces
    if (uiMode == 2){
        ui->sideView->setVisible( 0 );
    }

    ui->stackedWidget->setCurrentIndex( 2 );
	ui->toolsBar->setDisabled( true );
    connect(sMgr, SIGNAL(deviceAdded(QString)), this, SLOT(loadDrivePage()));
    connect(sMgr, SIGNAL(deviceRemoved(QString)), this, SLOT(loadDrivePage()));

    loadDrivePage();
}

void corefm::loadDrivePage()
{
    ui->driveView->clear();

    foreach (StorageDevice drive, sMgr->devices()) {
        foreach (StorageBlock block, drive.partitions()) {
            if (block.fileSystem().isEmpty() and block.property("Partition", "Number").toInt() == 0) {
                continue;
            }

            if (block.property("Partition", "IsContainer").toBool()) {
                continue;
            }

            QString name = block.label();

            if (name.isEmpty()) {
                name = block.path().split("/").takeLast();
            }
            name = name.trimmed();

            QString size = CPrime::FileFunc::formatSize(block.totalSize());
            QListWidgetItem *itemPartition = new QListWidgetItem(ui->driveView);
            itemPartition->setIcon(QIcon::fromTheme("drive-partition"));
            itemPartition->setText(name + " (" + size + ")");
            itemPartition->setData(Qt::UserRole + 1, block.path());
            ui->driveView->addItem(itemPartition);
            ui->driveView->setIconSize(listViewIconSize);
        }
    }
}

void corefm::openSelecteDrive(QListWidgetItem *item)
{
    QString blockPath = item->data(Qt::UserRole + 1).toString();
    if (blockPath.count()) {
        QString blockID = blockPath.split("/").takeLast();
        StorageBlock block(blockID);

        /* Try mounting anything that isn't swap */
        if (block.fileSystem() != "swap") {
            QString str = block.mountPoint().replace("\\040" , " ");
            if (block.mount()) {
                block = StorageBlock(blockID);
                str = block.mountPoint().replace("\\040" , " ");
                CPrime::InfoFunc::messageEngine("CoreFm", "corefm", "Mounted", "Partition '" + block.device() + "' mounted.", this);
            }
            ui->stackedWidget->setCurrentIndex(0);
            loadAddress(str);
        }
    }
}
